"""
    VoronoiFVM

$(read(joinpath(@__DIR__,"..","README.md"),String))
"""
module VoronoiFVM

using BandedMatrices: BandedMatrices, BandedMatrix, Zeros
using CommonSolve: CommonSolve, solve, solve!
using DiffResults: DiffResults
using DocStringExtensions: DocStringExtensions, SIGNATURES, TYPEDEF,
                           TYPEDFIELDS, TYPEDSIGNATURES
using ExtendableGrids: ExtendableGrids, BEdgeNodes, BFaceCells, BFaceEdges,
                       BFaceGeometries, BFaceNodes, BFaceNormals, BFaceRegions,
                       Cartesian1D, Cartesian2D, Cartesian3D, CellEdges,
                       CellGeometries, CellNodes, CellRegions,
                       CoordinateSystem, Coordinates, Cylindrical2D, Edge1D,
                       EdgeCells, EdgeNodes, ExtendableGrid,
                       Polar1D, Spherical1D, Tetrahedron3D,
                       Triangle2D, Vertex0D, VoronoiFaceCenters, coord_type,
                       dim_space, index_type, local_celledgenodes, num_bfaces,
                       num_cells, num_edges, num_nodes, num_cellregions, num_bfaceregions, num_targets,
                       simplexgrid, subgrid, tricircumcenter!

using ExtendableSparse: ExtendableSparse, BlockPreconditioner,
                        ExtendableSparseMatrix,
                        PointBlockILUZeroPreconditioner, factorize!, flush!,
                        nnz, rawupdateindex!, sparse, updateindex!

using ForwardDiff: ForwardDiff
using InteractiveUtils: InteractiveUtils
using JLD2: JLD2, jldopen
using LinearAlgebra: LinearAlgebra, Diagonal, I, Tridiagonal, isdiag, ldiv!, norm
using LinearSolve: LinearSolve, KLUFactorization, KrylovJL_BICGSTAB,
                   KrylovJL_CG, KrylovJL_GMRES, LinearProblem,
                   SparspakFactorization, UMFPACKFactorization, init
using Printf: Printf, @printf, @sprintf
using Random: Random, AbstractRNG
using RecursiveArrayTools: RecursiveArrayTools, AbstractDiffEqArray
using RecursiveFactorization: RecursiveFactorization
using SciMLBase: SciMLBase
using SparseArrays: SparseArrays, SparseMatrixCSC, dropzeros!, nonzeros,
                    nzrange, spzeros
using SparseDiffTools: SparseDiffTools, forwarddiff_color_jacobian!,
                       matrix_colors
using StaticArrays: StaticArrays, @MVector, @SArray, @SMatrix
using Statistics: Statistics, mean
using Symbolics: Symbolics
using Compat: @compat

include("vfvm_physics.jl")
@compat public Physics

include("vfvm_functions.jl")
export fbernoulli
export fbernoulli_pm
export inplace_linsolve!

include("vfvm_densesolution.jl")
include("vfvm_sparsesolution.jl")
export num_dof
export dof
export getdof
export setdof!

include("vfvm_history.jl")
export NewtonSolverHistory, TransientSolverHistory, details

include("vfvm_transientsolution.jl")
export TransientSolution

include("vfvm_xgrid.jl")
export cartesian!, circular_symmetric!, spherical_symmetric!
export coordinates

"""
$(TYPEDEF)
    
Abstract type for finite volume system structure.
"""
abstract type AbstractSystem{Tv <: Number, Tc <: Number, Ti <: Integer, Tm <: Integer} end
include("vfvm_geometryitems.jl")
include("vfvm_assemblydata.jl")
include("vfvm_system.jl")
export unknowns
export num_species
export enable_species!
export enable_boundary_species!
export update_grid!
export boundary_dirichlet!
export boundary_neumann!
export boundary_robin!
export ramp
export value
export physics!
export history, history_summary, history_details
export evaluate_residual_and_jacobian
export edgelength
export viewK, viewL, data
export hasoutflownode, isoutflownode, outflownode


@compat public System, AbstractSystem

# export to be deprecated
export partitioning, Equationwise

include("vfvm_formfactors.jl")
export meas, project
export unknown_indices
export edgevelocities, bfacevelocities, bfacenodefactors
export time, region, embedparam, parameters
export calc_divergences

include("vfvm_solvercontrol.jl")
export fixed_timesteps!, NewtonControl, SolverControl
include("vfvm_linsolve.jl")
export DirectSolver, GMRESIteration, CGIteration, BICGstabIteration, NoBlock, EquationBlock, PointBlock

include("vfvm_exceptions.jl")
include("vfvm_assembly.jl")
include("vfvm_solver.jl")
export solve!, solve

include("vfvm_postprocess.jl")
export nodeflux
export integrate
export l2norm, lpnorm
export w1pseminorm, h1seminorm
export w1pnorm, h1norm
export lpw1pnorm, l2h1norm
export lpw1pseminorm, l2h1seminorm
export nodevolumes

include("vfvm_testfunctions.jl")
export testfunction
export TestFunctionFactory

include("vfvm_quantities.jl")
export ContinuousQuantity
export DiscontinuousQuantity
export InterfaceQuantity
export subgrids, views

include("vfvm_impedance.jl")
export impedance, freqdomain_impedance
export measurement_derivative

include("vfvm_diffeq_interface.jl")
export eval_rhs!, eval_jacobian!, mass_matrix, prepare_diffeq!

include("gridvisualize.jl")
export plothistory
#include("precompile.jl")

end
